class Cat(val name: String){
  private val age = 10
}

//拡張関数。Catクラスの外にある関数で、Catクラスを拡張している
fun Cat.run(){
  //拡張関数の中では、Catクラスのプロパティを参照できる
  println("${this.name} is Running!")
  //エラー。privateプロパティは参照できない
  //println("${this.name}'s age is ${this.age}")
}

//エラー。Stringクラスはfinalなので継承できない
//class MyString: String(){}


//2つの数の和と積を返す関数。戻り値にPair<>を使うのがポイント
fun addAndMul(a: Int, b: Int): Pair<Int, Int>{
  //戻り値はPairオブジェクト
  return Pair(a + b, a * b)
}

//拡張関数。Stringクラスにdumpというメソッドを追加する（ように見せる）
fun String.dump(){
  println("Length: ${this.length}, Content: $this")
}

//データクラスは自動でcomponentNというメソッドが定義されているので分解宣言がそのまま使える
data class Person(val firstName: String, val lastName: String, val age: Int)

//自作の複素数クラス（整数のみ対応）
data class Complex(val real: Int, val image: Int){
  //operatorキーワードを使って+演算子をオーバーロード
  operator fun plus(value: Complex): Complex{
    return Complex(real + value.real, image + value.image)
  }
  //-演算子（2項）
  operator fun minus(value: Complex): Complex{
    return Complex(real - value.real, image - value.image)
  }
  //-演算子（単項）
  operator fun unaryMinus(): Complex{
    return Complex(-real, -image)
  }
  //複素数的な表記を返すように文字列化
  override fun toString(): String{
    if(image > 0)
      return "$real + ${image}i"
    else
      return "$real ${image}i"
  }
}

fun main(args:Array<String>) {

    //複素数クラスを使ってみる
    var c1 = Complex( 2, 2)  //  2 + 2i
    var c2 = Complex(-1, 5)  // -1 + 5i
    var c3 = c1 + c2 //普通に+演算子が使える

    println(c3)      //加算結果     「1 + 7i」
    println(c1 - c2) //減算結果     「3 -3i」
    println(-c1)     //単項の-演算子「-2 -2i」

    ///拡張関数 関連サンプル
    val cat = Cat("Tama") 
    //拡張関数の呼び出し例。Catクラスにrunメソッドが存在するように書ける
    cat.run() //「Tama is Running!」

    //Stringクラスに拡張関数として追加したdumpメソッドを呼ぶ
    "Hello".dump() //「Length: 5, Content: Hello」

    ///範囲 関連サンプル

    //1～10の範囲
    for(i in 1..10){
      print(i) //「12345678910」
    }
    println()

    //2～10の範囲で2刻み
    for(i in 2..10 step 2){
      print(i) //「246810」
    }
    println()

    //9～1の範囲（逆順）で2刻み
    for(i in 9 downTo 1 step 2){
      print(i) //「97531」
    }
    println()

    for(i in 9.downTo(1).step(2)){
      print(i) //「97531」
    }
    println()

    //1～10の範囲。ただし終わりの値（10）は含まない
    for(i in 1 until 10){
      print(i) //「123456789」
    }
    println()

    //a～zの範囲
    for(i in 'a'..'z'){
      print(i) // 「abcdefghijklmnopqrstuvwxyz」
    }
    println()

    //A～zの範囲で2刻み
    for(i in 'A' until 'z' step 2){
      print(i) // 「ACEGIKMOQSUWY[]_acegikmoqsuwy」
    }
    println()


    ///分解宣言と多重戻り値 関連サンプル
    //複数の戻り値を返す（ように見える関数）の呼び出し
    val (wa, seki) = addAndMul(2, 5)
    println("和: $wa, 積: $seki") //「和: 7, 積: 10」

    //実際には以下のような挙動
    val result = addAndMul(2, 5)
    val wa2 = result.component1()
    val seki2 = result.component2()
    println("和: $wa2, 積: $seki2") //「和: 7, 積: 10」

    //データクラスのインスタンスを作る
    val tanaka = Person("一郎", "田中", 25)
    //データクラスのオブジェクトを複数の変数に分解して代入する
    val (mei, sei, age) = tanaka
    println("姓: $sei, 名: $mei, 年齢: $age") //「姓: 田中, 名: 一郎, 年齢: 25」

    //分解宣言は変数名などは見ておらず、順番で制御されているので注意
    val (lastName, firstName, age2) = tanaka
    //lastName変数には1番目の値であるfirstNameプロパティ、firstName変数には2番目のlastNameプロパティの値が返る
    println("姓: $lastName, 名: $firstName, 年齢: $age2") //「姓: 一郎, 名: 田中, 年齢: 25」

    //受け取る変数の個数が少ない場合もエラーにはならない。受け側で足りないものは代入されない
    val (mei2, sei2) = tanaka
    println("姓: $sei2, 名: $mei2") //「姓: 田中, 名: 一郎」

    //受け取る変数の個数が多いとさすがにエラーになる
    //val (mei3, sei3, age3, other) = tanaka

    val names = arrayOf("Doi", "Yamada", "Sakurai")

    //配列からインデックスと要素の両方を取得する
    for((index, elem) in names.withIndex()){
      println("$index: $elem")
    }
}

